from django.db.models import Max
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from django.core.exceptions import PermissionDenied
from django.conf import settings
from rest_framework.generics import (
    GenericAPIView, ListAPIView, CreateAPIView, RetrieveAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView, RetrieveDestroyAPIView,
)
from rest_framework.views import APIView
from rest_framework.parsers import MultiPartParser, FileUploadParser
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError
from arkindex.project.mixins import CorpusACLMixin, SelectionMixin
from arkindex.project.permissions import IsVerified, IsAdminUser
from arkindex.project.openapi import AutoSchema
from arkindex.documents.models import Corpus, Right, Element
from arkindex.dataimport.models import \
    DataImport, DataFile, Repository, Event, EventType
from arkindex.dataimport.serializers.ml_tool import MLToolSerializer, MLToolTaskSerializer
from arkindex.dataimport.serializers.files import DataFileSerializer, DataFileCreateSerializer
from arkindex.dataimport.serializers.git import \
    RepositorySerializer, RepositoryStartImportSerializer, ExternalRepositorySerializer, EventSerializer
from arkindex.dataimport.serializers.imports import (
    DataImportLightSerializer, DataImportSerializer,
    DataImportFromFilesSerializer, ElementsWorkflowSerializer
)
from arkindex.users.models import OAuthCredentials
from arkindex_common.ml_tool import MLTool
from arkindex_common.enums import DataImportMode
from ponos.models import State
from datetime import datetime
import hashlib
import magic
import os.path


class DataImportsList(CorpusACLMixin, ListAPIView):
    """
    List data imports
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user)
        }

        if 'corpus' in self.request.query_params:
            filters['corpus'] = self.get_corpus(self.request.query_params['corpus'])

        if 'mode' in self.request.query_params:
            try:
                filters['mode'] = DataImportMode(self.request.query_params['mode'])
            except ValueError:
                raise ValidationError("Mode '{}' does not exist".format(self.request.query_params['mode']))

        if 'id' in self.request.query_params:
            filters['id__startswith'] = self.request.query_params['id']

        qs = DataImport.objects.filter(**filters).prefetch_related('workflow__tasks')

        # Order workflow by date of last updated task in workflow
        qs = qs.annotate(last_task=Max('workflow__tasks__updated'), last_run=Max('workflow__tasks__run'))
        return qs.order_by('-last_task')


class DataImportDetails(CorpusACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve and delete a data import
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        if instance.state == State.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataImportRetry(CorpusACLMixin, GenericAPIView):
    """
    Retry a failed import
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        if instance.workflow is not None and instance.state in (State.Unscheduled, State.Pending):
            raise ValidationError('This workflow is already pending')
        elif instance.state == State.Running:
            raise ValidationError('This workflow is already running')
        instance.retry()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class DataImportFromFiles(CreateAPIView):
    """
    Create, configure and start a workflow from one or more DataFiles
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFromFilesSerializer

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns a element for 'folder_id'
        folder = serializer.validated_data.get('folder_id')
        folder_name = serializer.validated_data.get('folder_name')
        folder_type_slug = serializer.validated_data.get('folder_type')
        element_type_slug = serializer.validated_data.get('element_type')
        pdf_engine = serializer.validated_data.get('pdf_engine')
        ml_tools = serializer.validated_data.get('ml_tools')

        if folder:
            # The files' corpus is already validated as writable
            if folder.corpus != corpus:
                raise ValidationError('Element and files are in different corpora')
        else:
            if not folder_name:
                if mode == DataImportMode.PDF:
                    suffix = os.path.splitext(files[0].name)[0]
                else:
                    suffix = datetime.today().strftime('%Y-%m-%d %H:%M')
                folder_name = 'Import {} {}'.format(mode.value, suffix)

            folder = corpus.elements.create(
                name=folder_name,
                type=corpus.types.get(slug=folder_type_slug),
            )
            folder.events.create(type=EventType.Addition)

        payload = {
            'folder_id': str(folder.id),
            'element_type': element_type_slug,
            # Re-serialize ML tools as JSON
            'ml_tools': MLToolTaskSerializer(ml_tools, many=True).data,
        }

        if mode == DataImportMode.PDF:
            assert pdf_engine is not None, 'PDF engine does not exist'
            payload['pdf_engine'] = pdf_engine.value

        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            payload=payload,
        )
        for f in files:
            self.dataimport.files.add(f)
        self.dataimport.start()


class CorpusWorkflow(SelectionMixin, CreateAPIView):
    """
    Start a Ponos ML workflow from Arkindex corpus elements
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementsWorkflowSerializer
    openapi_overrides = {
        'operationId': 'CreateElementsWorkflow',
        'description': 'Create a distributed workflow from elements of an Arkindex corpus',
        'tags': ['imports'],
        'responses': {
            '201': {
                'content': {
                    'application/json': {
                        'schema': AutoSchema()._map_serializer(DataImportSerializer())
                    }
                }
            }
        }
    }

    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        ml_tools, corpus, element, name, elt_type, selection = map(
            lambda key: serializer.validated_data.pop(key, None),
            ('ml_tools', 'corpus', 'element', 'name', 'type', 'selection')
        )

        payload = {
            'corpus_id': str(corpus.id),
            # Re-serialize ML tools as JSON
            'ml_tools': MLToolTaskSerializer(ml_tools, many=True).data,
            # Chunks and thumbnails parameters
            **serializer.validated_data
        }

        if selection:
            elements = self.get_selection(corpus.id)
            if not elements:
                raise ValidationError({
                    '__all__': ['No element match filtering']
                })
            payload['elements'] = list(map(str, elements.values_list('id', flat=True)))

        # Add filters
        if name:
            payload['name_contains'] = name
        if elt_type:
            payload['elt_type'] = elt_type
        if element:
            payload['element'] = str(element.id)
            # Add information concerning import
            if element.type.folder:
                payload['folder_id'] = str(element.id)
            else:
                try:
                    folder_elt = Element.objects.get_ascending(element.id, type__folder=True)[0]
                    payload['folder_id'] = str(folder_elt.id)
                except IndexError:
                    payload['folder_id'] = None

        dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Elements,
            payload=payload,
        )
        dataimport.start()
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(
                dataimport,
                context=self.get_serializer_context()
            ).data,
        )


class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk'])).prefetch_related('images__server')


class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        ).prefetch_related('images__server')

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class DataFileUpload(CorpusACLMixin, APIView):
    """
    Upload a new file to a corpus
    """
    permission_classes = (IsVerified, )
    parser_classes = (MultiPartParser, FileUploadParser)

    def post(self, request, pk=None, format=None):
        if 'file' not in request.FILES:
            raise ValidationError({'file': ['No file was sent in the request']})
        try:
            corpus = self.get_corpus(pk, right=Right.Write)
        except Http404:
            raise ValidationError({'corpus': ['Corpus not found']})

        file_obj = request.FILES['file']

        md5hash = hashlib.md5()
        for chunk in file_obj.chunks():
            md5hash.update(chunk)
        file_hash = md5hash.hexdigest()

        # Reopen file to reread from beginning
        file_obj.open()
        file_type = magic.from_buffer(file_obj.read(1024), mime=True)

        existing_file = corpus.files.filter(hash=file_hash).first()
        if existing_file:
            raise ValidationError({
                'file': ['File already exists'],
                'id': str(existing_file.id),
            })

        df = DataFile(
            hash=file_hash,
            corpus=corpus,
            name=file_obj.name,
            size=file_obj.size,
            content_type=file_type,
        )

        file_obj.open()
        df.s3_object.upload_fileobj(
            file_obj.file,
            ExtraArgs={'ContentType': df.content_type}
        )
        try:
            # If hash check is passed datafile instance will be saved
            df.check_hash(save=True, raise_exc=True)
        except ValueError:
            raise ValidationError({'file': ['Remote server hash do not match given file hash']})

        return Response(
            data=DataFileSerializer(df).data,
            status=status.HTTP_201_CREATED,
        )


class DataFileCreate(CreateAPIView):
    """
    Create a DataFile from user file hash
    A s3 PUT URI is returned to upload corresponding file
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer


class GitRepositoryImportHook(APIView):
    """
    Handle Git push events
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        assert repo.enabled, 'No credentials available'
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        )


class AvailableRepositoriesList(ListCreateAPIView):
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        return cred.git_provider_class(credentials=cred).list_repos(
            query=self.request.query_params.get('search'),
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])

        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(request=self.request, **serializer.validated_data)
        rev, _ = provider.get_or_create_latest_revision(repo)
        dataimport = rev.start_import()

        return Response(data={'import_id': str(dataimport.id)}, status=status.HTTP_201_CREATED)


class RepositoryRetrieve(CorpusACLMixin, RetrieveDestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class RepositoryStartImport(RetrieveAPIView):
    permission_classes = (IsAdminUser, )
    serializer_class = RepositoryStartImportSerializer

    def get_queryset(self):
        return Repository.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user),
        )

    def retrieve(self, request, *args, **kwargs):
        repo = self.get_object()

        if not repo.enabled:
            raise ValidationError("This repository does not have any associated credentials")

        rev, _ = repo.credentials.git_provider_class(
            credentials=repo.credentials,
        ).get_or_create_latest_revision(repo)

        di = rev.start_import()

        return Response(self.get_serializer(di).data)


class ElementHistory(ListAPIView):
    serializer_class = EventSerializer

    def get_queryset(self):
        return Event.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        )


class MLToolList(ListAPIView):
    """
    List available machine learning tools
    """
    serializer_class = MLToolSerializer
    pagination_class = None

    def get_queryset(self):
        return sorted(
            MLTool.list(settings.ML_CLASSIFIERS_DIR),
            key=lambda tool: tool.slug,
        )
