import os
import ijson
import logging
import requests
import urllib.parse
from io import BytesIO
from itertools import groupby
from django.conf import settings
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.db import transaction
from arkindex_common.enums import MetaType
from arkindex.documents.models import Corpus, Element, ElementType, MetaData
from arkindex.images.models import ImageServer, Zone
from arkindex.dataimport.models import Revision, EventType
from arkindex.project.polygon import Polygon
from arkindex.project.tools import random_string, get_or_instance


logger = logging.getLogger(__name__)


class IIIFParser(object):
    """
    Base IIIF parser
    """

    def __init__(self, path):
        self.path = path

    def open(self):
        """
        Open a stream for the given path/URL.
        """
        try:
            URLValidator(schemes=['http', 'https'])(self.path)
            resp = requests.get(self.path, timeout=settings.IIIF_DOWNLOAD_TIMEOUT)
            resp.raise_for_status()
            # Cannot use stream=True here: the parser uses the seek(int) method which is unsupported
            self.stream = BytesIO(resp.content)
            name, _ = os.path.splitext(os.path.basename(urllib.parse.urlparse(self.path).path))
            return name
        except ValidationError:
            assert os.path.isfile(self.path), "File does not exist"
            name, _ = os.path.splitext(os.path.basename(self.path))
            self.stream = open(self.path, 'rb')
            return name

    def close(self):
        """
        Close the stream if it is open.
        """
        if self.stream and not self.stream.closed:
            self.stream.close()

    def get_type(self):
        """
        Get a parser class for the given IIIF object.
        Raises ValueError if the object has no @type or if the type is not supported.
        """
        jsonld_type = next((
            value
            for prefix, event, value in ijson.parse(self.stream)
            if (prefix, event) == ('@type', 'string')
        ), None)

        if jsonld_type is None:
            raise ValueError("Missing @type property in JSON data")

        self.stream.seek(0)
        try:
            return PARSERS[jsonld_type]
        except KeyError:
            raise ValueError('Unsupported @type "{}"'.format(jsonld_type))

    @staticmethod
    def load(path, *args, **kwargs):
        p = IIIFParser(path)
        p.open()
        cls = p.get_type()
        p.close()
        return cls(path, *args, **kwargs)

    @staticmethod
    def from_dataimport(dataimport, path, config):
        p = IIIFParser(path)
        p.open()
        cls = p.get_type()
        p.close()
        return cls.from_dataimport(dataimport, path, config)


class ManifestParser(IIIFParser):
    """
    A class that parses a single IIIF manifest into a folder with pages and images.
    """

    def __init__(self, path, corpus,
                 revision=None, lazy=False, servers=[],
                 autocreate_servers=False, autoconvert_https=False,
                 folder_name=None, manifest_type=None, canvas_type=None,
                 **kwargs):
        """
        path: Path or URL to the IIIF manifest to parse.
        corpus: An instance of Corpus which will hold all the elements.
        revision: A Revision instance that should be used for change events on elements
        lazy: Boolean indicating whether the parser should not perform checks on image existence
        servers: An optional list of servers to restrict image server search
        autocreate_servers: If the ImageServer for a given image URL cannot be found,
                            will try to autodetect the server's URL and create new image servers
        autoconvert_https: If the ImageServer for a given image URL cannot be found and the URL uses HTTP,
                           will check the HTTPS version exists and try to match a server.
        folder_name: Set to a non-empty str to override the default folder name.
        manifest_type: An ElementType instance, or a slug for a folder type in the given corpus.
        canvas_type: An ElementType instance, or a slug for a normal type in the given corpus.
        """
        super().__init__(path)
        if revision:
            assert isinstance(revision, Revision)
        self.revision = revision
        assert isinstance(corpus, Corpus)
        self.corpus = corpus
        assert isinstance(lazy, bool)
        self.lazy = lazy
        if servers:
            assert isinstance(servers, list)
            self.servers = servers
        else:
            self.servers = list(ImageServer.objects.all())
        assert isinstance(autoconvert_https, bool)
        self.autoconvert_https = autoconvert_https
        assert isinstance(autocreate_servers, bool)
        self.autocreate_servers = autocreate_servers
        assert isinstance(manifest_type, ElementType)
        assert manifest_type.corpus == self.corpus and manifest_type.folder
        self.manifest_type = manifest_type
        assert isinstance(canvas_type, ElementType)
        assert canvas_type.corpus == self.corpus
        self.canvas_type = canvas_type
        self.stream = None

        # ImageServer that may get created on autocreate
        self.created_servers_count = 0

        self.folder_changed = False
        if folder_name:
            assert isinstance(folder_name, str)
        self.folder_name = folder_name

    @staticmethod
    def from_dataimport(dataimport, path, config):
        volumes_config = config['volumes']
        return ManifestParser(
            path,
            dataimport.corpus,
            revision=dataimport.revision,
            lazy=volumes_config['lazy_checks'],
            servers=volumes_config.image_servers,
            autocreate_servers=bool(volumes_config['image_servers']),
            autoconvert_https=volumes_config['autoconvert_https'],
            manifest_type=volumes_config['manifest_type'],
            canvas_type=volumes_config['canvas_type'],
        )

    def open(self):
        """
        Open a stream for the given manifest and save the folder name.
        """
        file_name = super().open()
        self.folder_name = self.folder_name or file_name

    def make_parents(self):
        self.folder, self.is_new = get_or_instance(
            Element,
            type=self.manifest_type,
            name=self.folder_name,
            corpus=self.corpus,
        )

    def find_image_server(self, image_url):
        try:  # Look in the specified servers
            return next(server for server in self.servers if image_url.startswith(server.url))
        except StopIteration:
            pass

        splat = urllib.parse.urlsplit(image_url)
        if self.autoconvert_https and splat.scheme == 'http':  # Try again but with HTTPS
            new_url = urllib.parse.urlunsplit(('https', *splat[1:]))
            logger.info("Trying '{}' instead of '{}'".format(new_url, image_url))
            try:
                # Look for a server before checking the server exists;
                # no need to wait for a HTTP request if there is no associated server
                serv = next(server for server in self.servers if new_url.startswith(server.url))
                requests.head(new_url, timeout=settings.IIIF_DOWNLOAD_TIMEOUT).raise_for_status()
                return serv
            except (requests.exceptions.RequestException, StopIteration):
                pass

        if not self.autocreate_servers:
            return

        logger.warning("No known image server for image {} - attempting autodetection".format(image_url))

        if splat.path.startswith('/iiif'):
            server_url = "{0}://{1}/iiif".format(splat.scheme, splat.netloc)
            server_name = '_'.join([self.corpus.name, random_string(5)])
            new_server = ImageServer.objects.create(display_name=server_name, url=server_url)
            self.servers.append(new_server)
            self.created_servers_count += 1

            logger.info("Created IIIF image server '{0}' with URL '{1}'".format(server_name, server_url))
            return new_server

    def parse_metadata(self):
        """
        Parse a manifest's metadata property into MetaData elements linked to the folder.
        """
        logger.info("Parsing metadata")
        self.metadata = []

        data = []
        for item in ijson.items(self.stream, 'metadata.item'):
            if not all(prop in item for prop in ('label', 'value')):
                logger.warning('Metadata does not have the required label and value properties')
                continue
            if not item['value']:
                logger.warning('Metadata with label {} has an empty value - ignoring'.format(item['label']))
                continue
            data.append(item)

        def grouper(item):
            return item['label']

        for label, items in groupby(sorted(data, key=grouper), key=grouper):
            for index, item in enumerate(items):
                md, created = get_or_instance(
                    MetaData,
                    element=self.folder,
                    type=MetaType.Text,
                    name=label,
                    index=index,
                    defaults={
                        'revision': self.revision,
                        'value': item['value'],
                    }
                )
                # Set folder as changed whenever any metadata is created
                self.folder_changed = self.folder_changed or created
                # Only update existing metadata revisions when values get updated
                if not created and md.value != item['value']:
                    # Set folder as changed whenever any metadata is updated
                    self.folder_changed = True
                    md.value = item['value']
                    md.revision = self.revision
                self.metadata.append(md)

        self.stream.seek(0)

    def parse_canvases(self):
        """
        Parse all canvases in all sequences of the manifest into Page
        """
        logger.info("Parsing canvases")
        self.change_count, self.pages, self.images = 0, [], []
        for i, canvas in enumerate(ijson.items(self.stream, 'sequences.item.canvases.item')):
            try:
                page = self.parse_canvas(canvas)
                logger.debug('Parsed page {}'.format(page.name))
            except Exception as e:
                logger.warning('Failed parsing canvas #{} ({}): {}'.format(i, canvas.get('@id'), str(e)), exc_info=1)
                continue

    def parse_canvas(self, canvas):
        """
        Parse a single canvas into a Page and perform diff
        """
        # Label contains the folio
        folio = canvas.get('label')
        if folio is None:
            raise ValueError("No folio (label) specified")

        if 'images' not in canvas or len(canvas['images']) < 1:
            raise ValueError("Canvas has no image")

        # Get the image resource
        resource = canvas['images'][0].get('resource')
        if resource is None:
            raise ValueError("Canvas has no image resource")

        # Go find the service ID to get the image URL
        if 'service' not in resource:
            raise ValueError("Image resource with ID {} has no IIIF image service".format(resource.get('@id')))
        if '@id' not in resource['service']:
            raise ValueError("Image service on resource {} has no @id".format(resource.get('@id')))
        service_id = resource['service']['@id']

        # Find the right server
        image_server = self.find_image_server(service_id)
        if image_server is None:
            raise ValueError("No image server found for image {}".format(service_id))

        # Strip server URL to get just the path
        image_path = service_id[len(image_server.url):].lstrip('/')
        image = image_server.find_image(
            image_path,
            offline=self.lazy,
            width=canvas['width'],
            height=canvas['height'],
        )
        self.images.append(image)

        poly = Polygon.from_coords(0, 0, canvas['width'], canvas['height'])
        zone, created = get_or_instance(
            Zone,
            image=image,
            polygon=poly,
        )

        new_page = Element(
            type=self.canvas_type,
            zone=zone,
            corpus_id=self.corpus.id,
            name=folio,
        )

        self.change_count += 1
        self.pages.append(new_page)
        return new_page

    def parse(self):
        """
        Run the full parsing process.
        """
        self.open()
        self.make_parents()
        self.parse_metadata()
        self.parse_canvases()

    @transaction.atomic
    def save(self):
        logger.info('Saving folder...')
        # There is no point of comparison between folders if their names changed;
        # we can only recreate them.
        self.folder.save()

        # Volume metadata can on the other hand get updated or deleted
        logger.info('Saving metadata...')
        for md in self.metadata:
            md.save()
        deleted_metadatas = self.folder.metadatas.exclude(id__in=[md.id for md in self.metadata])
        if deleted_metadatas.exists():
            self.folder_changed = True
            deleted_metadatas.delete()

        if self.images:
            logger.info('Saving images...')
            for image in self.images:
                image.save()

        if self.pages:
            logger.info('Saving changed pages...')
            for i, page in enumerate(self.pages):
                # Set folder as changed since a page changed
                self.folder_changed = True
                page.zone.save()
                page.save()
                page.events.create(
                    revision=self.revision,
                    type=EventType.Addition,
                )
                # Create path or update ordering
                page.add_parent(self.folder, order=i)
                page.metadatas.create(
                    name='folio',
                    type=MetaType.Text,
                    value=page.name,
                )

        if not self.is_new:  # Deleted pages cannot happen on a new folder
            deleted_pages = Element.objects \
                .get_descending(self.folder.id) \
                .filter(type=self.canvas_type) \
                .exclude(id__in=[p.id for p in self.pages])
            if deleted_pages.exists():
                # Set folder as changed since one or more pages got deleted
                self.folder_changed = True
                logger.info('Removing {} deleted pages...'.format(len(deleted_pages)))
                deleted_pages.delete()

        if self.is_new or self.folder_changed:
            self.folder.events.get_or_create(
                revision=self.revision,
                defaults={'type': EventType.Addition if self.is_new else EventType.Edit},
            )

    def run(self):
        try:
            self.parse()
            logger.info("Parsed folder {}: {} metadata in folder, {} pages ({} changed), {} new servers".format(
                self.folder.name,
                len(self.metadata), len(self.pages), self.change_count, self.created_servers_count
            ))
            self.save()
        finally:
            self.close()

        return self.pages + [self.folder]


class CollectionParser(IIIFParser):
    """
    A class that runs multiple ManifestParsers from a IIIF collection.
    """

    def __init__(self, path, *args, dataimport=None, config=None, **kwargs):
        super().__init__(path)

        self.dataimport = dataimport
        self.config = config
        if self.dataimport and not self.config or self.config and not self.dataimport:
            raise ValueError('CollectionParser requires both a DataImport and Config, or none of them')

        self.manifest_parser_args = args
        self.manifest_parser_kwargs = kwargs

        self.elements = []

    def _list_from(self, prop):
        self.stream.seek(0)
        items = []
        for data in ijson.items(self.stream, '{}.item'.format(prop)):
            if data.get('@type') != 'sc:Manifest':
                logger.warning('Unsupported type {} found in collection {}'.format(data.get('@type'), self.path))
                continue
            items.append((data['label'], data['@id']))
        return items

    def list(self):
        # IIIF 3: try using members; fallback to manifests + collections
        items = self._list_from('members')
        if not items:
            items = self._list_from('collections') + self._list_from('manifests')
        return items

    def run(self):
        try:
            self.open()
            for folder_name, manifest_path in self.list():
                try:
                    if self.dataimport:
                        parser = ManifestParser.from_dataimport(
                            self.dataimport,
                            manifest_path,
                            self.config,
                        )
                    else:
                        parser = ManifestParser(
                            manifest_path,
                            *self.manifest_parser_args,
                            **self.manifest_parser_kwargs,
                        )
                    # Override the folder name with the label from the collection
                    parser.folder_name = folder_name

                    logger.info('Parsing folder {}'.format(folder_name))
                    self.elements.extend(parser.run())
                except Exception as e:
                    logger.error('Failed importing manifest "{}" from collection "{}": {}'.format(
                        manifest_path, self.path, str(e)))
            return self.elements
        finally:
            self.close()

    @staticmethod
    def from_dataimport(dataimport, path, config):
        return CollectionParser(path, dataimport=dataimport, config=config)


PARSERS = {
    'sc:Manifest': ManifestParser,
    'sc:Collection': CollectionParser,
}
