from rest_framework import serializers
from rest_framework.utils import model_meta
from django.db.models import Q
from ponos.models import State
from arkindex_common.enums import DataImportMode, DataImportPDFEngine
from arkindex.project.serializer_fields import EnumField
from arkindex.documents.models import Corpus, Element
from arkindex.dataimport.models import DataImport, DataFile
from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.dataimport.serializers.ml_tool import MLToolTaskSerializer


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(State)
    mode = EnumField(DataImportMode)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    class Meta:
        model = DataImport
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
        )
        read_only_fields = ('id', 'state')


class ImagesPayloadSerializer(serializers.Serializer):
    """
    Serialize an image importing payload
    """

    folder_id = serializers.UUIDField()
    element_type = serializers.SlugField()
    ml_tools = MLToolTaskSerializer(many=True, required=False, default=list)


class PDFPayloadSerializer(ImagesPayloadSerializer):
    """
    Serialize a pdf importing payload
    """

    pdf_engine = EnumField(DataImportPDFEngine, default=DataImportPDFEngine.Convert)


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its payload
    """

    # Redefine state as writable
    state = EnumField(State)
    payload = serializers.JSONField()
    revision = RevisionSerializer(read_only=True)
    workflow = serializers.HyperlinkedRelatedField(read_only=True, view_name='ponos:workflow-details')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance is None:
            return
        payload_serializers = {
            DataImportMode.Images: ImagesPayloadSerializer,
            DataImportMode.PDF: PDFPayloadSerializer,
        }
        self.fields['payload'] = payload_serializers.get(self.instance.mode, serializers.JSONField)()

    def update(self, instance, validated_data):
        # ModelSerializer disallows updating with nested serializers.
        # We dynamically inject a payload serializer, but it is useless for the database
        # since the payload field is a JSONField with no particular checks.
        # We don't have to care about a nested serializer and we can treat it just as a dict.
        # This code is copy-pasted from DRF's ModelSerializer.
        info = model_meta.get_field_info(instance)
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)
        instance.save()
        return instance

    class Meta(DataImportLightSerializer.Meta):
        fields = DataImportLightSerializer.Meta.fields + (
            'payload',
            'files',
            'revision',
            'workflow',
        )


class DataImportFromFilesSerializer(serializers.Serializer):

    mode = EnumField(DataImportMode, default=DataImportMode.Images)
    files = serializers.PrimaryKeyRelatedField(queryset=DataFile.objects.all(), many=True)
    folder_id = serializers.UUIDField(required=False, allow_null=True)
    folder_name = serializers.CharField(max_length=250, required=False, allow_null=True)
    folder_type = serializers.SlugField(required=False, allow_null=True)
    element_type = serializers.SlugField()
    pdf_engine = EnumField(DataImportPDFEngine, default=DataImportPDFEngine.Convert)
    ml_tools = MLToolTaskSerializer(many=True, required=False, allow_null=True)

    default_error_messages = {
        'mode_not_allowed': 'This mode is not allowed when importing from files',
        'files_required': 'At least one file is required to start an import',
        'unique_corpus': 'Imports can only run on files from a single corpus',
        'corpus_read_only': 'Cannot write in corpus',
        'folder_not_found': 'Folder does not exist',
        'unique_pdf': 'Cannot import more than one PDF at once',
        'pdf_only': 'PDF imports can only import PDF files',
        'images_only': 'Image imports can only import image files',
        'folder_required': 'Specify either folder_type or folder_id',
        'type_not_found': 'Element type {slug!r} does not exist',
        'type_not_folder': 'Element type {slug!r} is not a folder',
        'type_folder': 'Element type {slug!r} is a folder',
        'type_hidden': 'Element type {slug!r} is hidden',
    }

    def validate_mode(self, mode):
        if mode not in (DataImportMode.Images, DataImportMode.PDF):
            self.fail('mode_not_allowed')
        return mode

    def validate_files(self, files):
        if not files:
            self.fail('files_required')
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            self.fail('unique_corpus')
        corpus = corpora.pop()
        if corpus not in Corpus.objects.writable(self.context['request'].user):
            self.fail('corpus_read_only')
        return files

    def validate_folder_id(self, folder_id):
        if not folder_id:
            return None
        try:
            return Element.objects.get(
                id=folder_id,
                type__folder=True,
            )
        except Element.DoesNotExist:
            self.fail('folder_not_found')

    def validate(self, data):
        if data['mode'] == DataImportMode.PDF:
            if len(data['files']) > 1:
                self.fail('unique_pdf')
            if data['files'][0].content_type != 'application/pdf':
                self.fail('pdf_only')

        elif data['mode'] == DataImportMode.Images:
            if not all(f.content_type.startswith('image/') for f in data['files']):
                self.fail('images_only')

        else:
            raise NotImplementedError

        element_type_slug = data.get('element_type')
        element_type = data['files'][0].corpus.types.filter(slug=element_type_slug).first()
        if not element_type:
            self.fail('type_not_found', slug=element_type_slug)
        if element_type.folder:
            self.fail('type_folder', slug=element_type_slug)
        if element_type.hidden:
            self.fail('type_hidden', slug=element_type_slug)

        folder_type_slug, folder_id = data.get('folder_type'), data.get('folder_id')

        if not (bool(folder_type_slug) ^ bool(folder_id)):
            self.fail('folder_required')

        if folder_type_slug:
            folder_type = data['files'][0].corpus.types.filter(slug=folder_type_slug).first()
            if not folder_type:
                self.fail('type_not_found', slug=folder_type_slug)
            if not folder_type.folder:
                self.fail('type_not_folder', slug=folder_type_slug)
            if folder_type.hidden:
                self.fail('type_hidden', slug=folder_type_slug)

        return data


class ElementsWorkflowSerializer(serializers.Serializer):

    ml_tools = MLToolTaskSerializer(many=True)
    chunks = serializers.IntegerField(min_value=1, max_value=10, default=1)
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    thumbnails = serializers.BooleanField(required=False)

    # Elements filtering options
    element = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none(), required=False)
    name = serializers.CharField(required=False)
    type = serializers.SlugField(required=False)
    selection = serializers.BooleanField(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora
        # Only elements representing a folder or containing a zone can be imported
        self.fields['element'].queryset = Element.objects.filter(
            Q(type__folder=True) | Q(zone__isnull=False),
            corpus__in=corpora
        )

    def validate(self, data):
        """
        Handle filters validation
        """
        corpus = data.get('corpus')
        targeted_elts = corpus.elements

        element = data.get('element')
        name = data.get('name')
        elt_type = data.get('type')
        selection = data.get('selection')

        if selection and (element or name or elt_type):
            raise serializers.ValidationError({
                '__all__': ['Filtering parameters (element, name, type) cannot be used with selection parameter']
            })
        if element and (name or elt_type):
            raise serializers.ValidationError({
                '__all__': ['Filtering parameters (name, type) cannot be used with element parameter']
            })
        if element:
            # Assert element is in the right corpus
            if corpus.id != element.corpus.id:
                raise serializers.ValidationError({
                    'elements': ['Element is not part of corpus {}'.format(corpus.name)]
                })
        if elt_type:
            # Check type filter is valid
            if not corpus.types.filter(slug=elt_type).exists():
                raise serializers.ValidationError({
                    'type': ['Unrecognized element type filter']
                })
            targeted_elts = targeted_elts.filter(type__slug=elt_type)

        if name:
            targeted_elts = targeted_elts.filter(name__icontains=name)

        # Assert filters match Arkindex elements
        if not targeted_elts.exists():
            raise serializers.ValidationError({
                '__all__': ['No element match filtering']
            })
        return data
