from unittest.mock import patch, call, MagicMock
from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import DataImportMode, DataImportPDFEngine
from arkindex.dataimport.models import DataImport, DataFile
from arkindex.documents.models import Element, Corpus
from arkindex.project.tests import FixtureAPITestCase
from ponos.models import State
import uuid


class TestImports(FixtureAPITestCase):
    """
    Test data imports management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.img_df = DataFile.objects.create(
            name='test.jpg', size=42, hash='aaaa', content_type='image/jpg', corpus=cls.corpus)
        cls.pdf_df = DataFile.objects.create(
            name='test.pdf', size=42, hash='bbbb', content_type='application/pdf', corpus=cls.corpus)

    def setUp(self):
        super().setUp()
        self.dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Images)

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        result = data['results'][0]
        self.assertEqual(result['id'], str(self.dataimport.id))
        self.assertEqual(result['state'], State.Unscheduled.value)
        self.assertEqual(result['mode'], DataImportMode.Images.value)
        self.assertEqual(result['corpus'], str(self.corpus.id))

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        dataimport2 = corpus2.imports.create(creator=self.user, mode=DataImportMode.Images)

        response = self.client.get(reverse('api:import-list'), {'corpus': str(corpus2.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        dataimport2 = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'mode': DataImportMode.PDF.value})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first DataImport
        dataimport_id = str(uuid.uuid4())
        while dataimport_id[:10] == str(self.dataimport.id)[:10]:
            dataimport_id = str(uuid.uuid4())

        dataimport2 = self.corpus.imports.create(
            id=dataimport_id,
            creator=self.user,
            mode=DataImportMode.Images,
        )
        response = self.client.get(reverse('api:import-list'), {'id': dataimport_id[:10]})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['state'], State.Unscheduled.value)

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_no_permission(self):
        self.client.force_login(self.user)
        acl_rights = self.corpus.corpus_right.get(user=self.user)
        acl_rights.can_write = False
        acl_rights.can_admin = False
        acl_rights.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_details_destroy_verified(self):
        self.user.verified_email = False
        self.user.save()
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_details_destroy(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_details(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.dataimport.id))

    def test_details_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.dataimport.workflow)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().delete()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['state'], State.Unscheduled.value)

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry_running(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        task = self.dataimport.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry(self):
        self.client.force_login(self.user)
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)

    def test_retry_no_workflow(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.dataimport.workflow)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataimport.refresh_from_db()
        self.assertEqual(self.dataimport.state, State.Unscheduled)
        self.assertIsNotNone(self.dataimport.workflow)

    def test_retry_repo_disabled(self):
        self.client.force_login(self.user)
        self.dataimport.mode = DataImportMode.Repository
        self.dataimport.revision = self.rev
        self.dataimport.save()
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.dataimport.state, State.Error)
        self.creds.delete()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.creds.save()

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertIn('folder_id', dataimport.payload)
        self.assertIn('element_type', dataimport.payload)
        self.assertNotIn('classifier', dataimport.payload)
        self.assertNotIn('recognizer', dataimport.payload)
        self.assertListEqual(dataimport.ml_tools, [])

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertEqual(Element.objects.get(id=dataimport.payload['folder_id']).name, 'Import pdf test')

    def test_from_files_pdf_convert(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'pdf_engine': 'convert',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertEqual(dataimport.pdf_engine, DataImportPDFEngine.Convert)
        self.assertEqual(Element.objects.get(id=dataimport.payload['folder_id']).name, 'Import pdf test')

    def test_from_files_pdf_poppler(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'pdf_engine': 'poppler',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertEqual(dataimport.pdf_engine, DataImportPDFEngine.Poppler)
        self.assertEqual(Element.objects.get(id=dataimport.payload['folder_id']).name, 'Import pdf test')

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_single(self):
        self.client.force_login(self.user)
        pdf2 = DataFile.objects.create(
            name='test2.pdf', size=1337, hash='0', content_type='application/pdf', corpus=self.corpus)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id), str(pdf2.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'pdf',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_images_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'folder_id': str(volume.id),
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.payload['folder_id'], str(volume.id))

    def test_from_files_folder_name(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'folder_name': 'New volume',
                'folder_type': 'volume',
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        volume = Element.objects.get(id=dataimport.payload['folder_id'])
        self.assertEqual(volume.name, 'New volume')
        self.assertEqual(volume.corpus, self.corpus)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex_common.ml_tool.MLTool.get')
    def test_from_files_set_tools(self, ml_get_mock):
        classifier_mock, recognizer_mock = MagicMock(), MagicMock()
        classifier_mock.type, classifier_mock.slug = MLToolType.Classifier, 'someclassifier'
        recognizer_mock.type, recognizer_mock.slug = MLToolType.Recognizer, 'somerecognizer'
        classifier_mock.get_task_recipe.return_value = {'image': 'some_image1', 'command': 'some_command1'}
        recognizer_mock.get_task_recipe.return_value = {'image': 'some_image2', 'command': 'some_command2'}
        tools = {'someclassifier': classifier_mock, 'somerecognizer': recognizer_mock}
        ml_get_mock.side_effect = lambda directory, type, slug: tools[slug]

        self.client.force_login(self.user)

        with self.settings(ML_CLASSIFIERS_DIR='/somewhere'):
            response = self.client.post(reverse('api:import-from-files'), {
                'files': [str(self.img_df.id)],
                'folder_type': 'volume',
                'element_type': 'page',
                'ml_tools': [
                    {'type': 'classifier', 'slug': 'someclassifier'},
                    {
                        'type': 'recognizer',
                        'slug': 'somerecognizer',
                        'parents': [{'type': 'classifier', 'slug': 'someclassifier'}],
                    },
                ],
            }, format='json')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertListEqual(dataimport.payload.get('ml_tools'), [
            {'type': 'classifier', 'slug': 'someclassifier', 'parents': []},
            {
                'type': 'recognizer',
                'slug': 'somerecognizer',
                'parents': [{'type': 'classifier', 'slug': 'someclassifier'}],
            },
        ])
        self.assertEqual(ml_get_mock.call_count, 6)
        self.assertListEqual(ml_get_mock.call_args_list, [
            call('/somewhere', MLToolType.Classifier, 'someclassifier'),
            call('/somewhere', MLToolType.Classifier, 'someclassifier'),
            call('/somewhere', MLToolType.Recognizer, 'somerecognizer'),
        ] * 2)
