from arkindex.documents.models import EntityRole, Entity, Corpus, Element, MLClass, Transcription, TranscriptionEntity,\
    Right, EntityLink, DataSource, ClassificationState
from arkindex.project.mixins import CorpusACLMixin
from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, RetrieveAPIView, CreateAPIView, RetrieveUpdateDestroyAPIView
from arkindex.documents.serializers.entities import (
    EntityRoleSerializer, EntitySerializer, EntityCreateSerializer, EntityLinkCreateSerializer,
    TranscriptionEntitySerializer, TranscriptionEntityDetailsSerializer, ElementEntitiesSerializer,
    EntityLinkSerializer,
)
from arkindex.documents.serializers.elements import ElementSlimSerializer
from rest_framework import serializers, status
from rest_framework.response import Response
from django.core.exceptions import ValidationError
from arkindex.documents.serializers.ml import CountMLClassSerializer, MLClassSerializer
from arkindex.project.mixins import NestedCorpusMixin
from arkindex.project.permissions import IsVerified
from django.db.models import Count, Q


class CorpusRoles(CorpusACLMixin, ListCreateAPIView):
    """
    List all roles in a corpus
    """
    serializer_class = EntityRoleSerializer

    def get_queryset(self):
        return EntityRole.objects \
            .filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .order_by('parent_name', 'child_name')

    def perform_create(self, serializer):
        data = self.request.data
        if EntityRole.objects.filter(
            parent_name=data['parent_name'],
            child_name=data['child_name'],
            parent_type=data['parent_type'],
            child_type=data['child_type'],
            corpus_id=self.request.parser_context['kwargs']['pk']
        ).exists():
            raise serializers.ValidationError({
                'corpus': ['Role already exists in this corpus'],
                'id': self.request.parser_context['kwargs']['pk']
            })
        super().perform_create(serializer)


class CorpusMLClassList(NestedCorpusMixin, ListAPIView):
    """
    List available classes in a corpus with their distribution over elements
    """
    serializer_class = CountMLClassSerializer
    openapi_overrides = {
        'operationId': 'ListCorpusMLClasses',
        'description': (
            'List all available classes in a corpus with their'
            ' distribution over elements (best classes count)'
        ),
        'tags': ['ml'],
    }

    def get_queryset(self):
        if not hasattr(self, 'corpus'):  # OpenAPI support
            return MLClass.objects.none()
        best_classification_filter = Q(
            # Keep non rejected best or validated classifications
            (
                (
                    Q(classifications__high_confidence=True)
                    & ~Q(classifications__state=ClassificationState.Rejected)
                )
                | Q(classifications__state=ClassificationState.Validated)
            )
        )
        return MLClass.objects.filter(corpus=self.corpus) \
            .annotate(nb_best=Count(
                'classifications',
                filter=best_classification_filter,
            ))


class MLClassList(ListAPIView):
    """
    List all available classes
    """
    serializer_class = MLClassSerializer
    queryset = MLClass.objects.all()
    openapi_overrides = {
        'operationId': 'ListMLClasses',
        'description': 'List available machine learning classes in all corpora.',
        'tags': ['ml'],
    }


class EntityDetails(RetrieveUpdateDestroyAPIView):
    """
    Get all information about entity
    """
    serializer_class = EntitySerializer

    def get_queryset(self):
        return Entity.objects \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related(
                'parents__role',
                'children__role',
                'parents__child',
                'parents__parent',
                'children__parent',
                'children__child',
                'corpus'
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method not in ('GET', 'OPTIONS') and Right.Write not in obj.corpus.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')


class EntityElements(ListAPIView):
    """
    Get all elements that have a link with the entity
    """
    serializer_class = ElementSlimSerializer

    def get_queryset(self):
        pk = self.kwargs['pk']
        metadata_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                metadatas__entity_id=pk
            ).select_related('type')
        transcription_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                transcriptions__transcription_entities__entity_id=pk
            ).select_related('type')
        return metadata_elements.union(transcription_elements) \
            .order_by('name', 'type') \
            .prefetch_related('metadatas__entity', 'metadatas__revision', 'corpus')


class EntityCreate(CreateAPIView):
    """
    Create a new entity
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityCreateSerializer

    def perform_create(self, serializer):
        name = serializer.validated_data['name']
        type = serializer.validated_data['type']
        corpus = serializer.validated_data['corpus']
        metas = serializer.validated_data['metas'] if 'metas' in serializer.data else None
        source = DataSource.from_ml_tool(serializer.validated_data['ner'])
        return Entity.objects.create(
            name=name,
            type=type,
            corpus=corpus,
            metas=metas,
            source=source
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        entity = EntitySerializer(self.perform_create(serializer))
        headers = self.get_success_headers(serializer.data)
        return Response(entity.data, status=status.HTTP_201_CREATED, headers=headers)


class EntityLinkCreate(CreateAPIView):
    """
    Create a new link
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityLinkCreateSerializer


class TranscriptionEntityCreate(CreateAPIView):
    """
    Link an entity to a transcription
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionEntitySerializer

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if not self.request:
            # OpenAPI cannot pass request context
            return context
        tr = Transcription.objects.filter(
            id=self.kwargs.get('pk'),
            element__corpus__in=Corpus.objects.writable(self.request.user)
        ).first()
        return {**context, 'transcription': tr}

    def perform_create(self, serializer):
        data = serializer.validated_data
        if TranscriptionEntity.objects.filter(**data).exists():
            raise serializers.ValidationError({
                '__all__': ['Entity is already linked to this transcription at this position']
            })
        try:
            TranscriptionEntity(**data).full_clean()
        except ValidationError as e:
            raise serializers.ValidationError(serializers.as_serializer_error(e))
        super().perform_create(serializer)


class TranscriptionEntities(ListAPIView):
    """
    Retrieve all entities linked to a specific transcription
    """
    serializer_class = TranscriptionEntityDetailsSerializer

    def get_queryset(self):
        return TranscriptionEntity.objects \
            .filter(
                transcription_id=self.kwargs['pk'],
                transcription__element__corpus__in=Corpus.objects.readable(self.request.user)
            ) \
            .order_by('offset') \
            .prefetch_related('entity')


class ElementEntities(RetrieveAPIView):
    """
    List all entities linked to an element's transcriptions and metadata
    """
    serializer_class = ElementEntitiesSerializer

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        return Element.objects.filter(corpus__in=corpora).prefetch_related('type')


class ElementLinks(ListAPIView):
    """
    List all links where parent and child are linked to the element
    """
    serializer_class = EntityLinkSerializer

    def get_queryset(self):
        element = Element.objects.get(id=self.kwargs['pk'])
        entities = Entity.objects.filter(Q(metadatas__element_id=element.id) | Q(transcriptions__element_id=element.id))
        return EntityLink.objects.filter(parent__in=entities, child__in=entities) \
            .prefetch_related('role', 'child', 'parent').order_by('parent__name')
