from unittest.mock import patch
from asyncmock import AsyncMock
from django.db.models import Q
from arkindex_common.enums import TranscriptionType, MetaType, EntityType
from arkindex.project.polygon import Polygon
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.consumers import MLResultsConsumer
from arkindex.documents.models import Element, Transcription, RegionElement, Entity, Classification, DataSource


class TestMLResultsConsumer(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.folder1 = cls.corpus.elements.get(name='Volume 1')
        cls.folder2 = cls.corpus.elements.get(name='Volume 2')

        source = DataSource.objects.get(slug='test')
        ml_class = cls.corpus.ml_classes.create(name='Some class')

        cls.page1 = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.page1.classifications.create(
            ml_class=ml_class,
            confidence=0.42,
            source=source,
        )
        cls.page1.regions.create(
            zone=cls.page1.zone.image.zones.create(polygon=Polygon.from_coords(10, 10, 10, 10)),
            source=source,
            ml_class=ml_class,
        )
        entity1 = cls.corpus.entities.create(
            type=EntityType.Person,
            name='Some entity 1',
            source=source,
        )
        cls.page1.metadatas.create(
            type=MetaType.Text,
            name='something',
            value='Some entity 1',
            entity=entity1,
        )
        cls.page1.transcriptions.first().transcription_entities.create(
            offset=0,
            length=1,
            entity=entity1,
        )

        cls.page2 = cls.corpus.elements.get(name='Volume 2, page 1r')
        cls.page2.classifications.create(
            ml_class=ml_class,
            confidence=0.42,
            source=source,
        )
        cls.page2.regions.create(
            zone=cls.page2.zone.image.zones.create(polygon=Polygon.from_coords(10, 10, 10, 10)),
            source=source,
            ml_class=ml_class,
        )
        entity2 = cls.corpus.entities.create(
            type=EntityType.Person,
            name='Some entity 2',
            source=source,
        )
        cls.page2.metadatas.create(
            type=MetaType.Text,
            name='something',
            value='Some entity 2',
            entity=entity2,
        )
        cls.page2.transcriptions.create(
            score=0.74,
            text='some text',
            type=TranscriptionType.Line,
            source=source,
        ).transcription_entities.create(
            offset=0,
            length=1,
            entity=entity2,
        )

    def _get_querysets(self, elements):
        return [
            Transcription.objects.filter(element__in=elements),
            RegionElement.objects.filter(element__in=elements),
            Entity.objects.filter(transcriptions__element__in=elements),
            Entity.objects.filter(metadatas__element__in=elements),
            Classification.objects.filter(element__in=elements),
        ]

    def test_delete_missing_parameters(self):
        with self.assertRaises(AssertionError):
            MLResultsConsumer({}).ml_results_delete({})

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_delete_corpus(self, get_layer_mock):
        get_layer_mock.return_value.send = AsyncMock()
        querysets = self._get_querysets(self.corpus.elements.all())

        for queryset in querysets:
            self.assertTrue(queryset.exists())

        self.assertEqual(self.page1.metadatas.count(), 2)
        self.assertEqual(self.page2.metadatas.count(), 2)

        with self.assertNumQueries(16):
            MLResultsConsumer({}).ml_results_delete({'corpus_id': str(self.corpus.id)})

        for queryset in querysets:
            self.assertFalse(queryset.exists())

        self.assertEqual(self.page1.metadatas.count(), 2)
        self.assertEqual(self.page2.metadatas.count(), 2)
        metadata = self.page1.metadatas.get(value='Some entity 1')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertIsNone(metadata.entity)
        metadata = self.page2.metadatas.get(value='Some entity 2')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertIsNone(metadata.entity)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': None,
            'corpus': str(self.corpus.id),
            'transcriptions': True,
            'elements': True,
            'entities': True,
            'drop': False,
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_delete_element(self, get_layer_mock):
        get_layer_mock.return_value.send = AsyncMock()
        folder1_querysets = self._get_querysets(
            Element.objects.filter(Q(id=self.folder1.id) | Q(paths__path__last=self.folder1.id))
        )
        folder2_querysets = self._get_querysets(
            Element.objects.filter(Q(id=self.folder2.id) | Q(paths__path__last=self.folder2.id))
        )

        for queryset in folder1_querysets:
            self.assertTrue(queryset.exists())
        for queryset in folder2_querysets:
            self.assertTrue(queryset.exists())

        with self.assertNumQueries(17):
            MLResultsConsumer({}).ml_results_delete({'element_id': str(self.folder1.id)})

        for queryset in folder1_querysets:
            self.assertFalse(queryset.exists())
        for queryset in folder2_querysets:  # This folder should not change
            self.assertTrue(queryset.exists())

        self.assertEqual(self.page1.metadatas.count(), 2)
        self.assertEqual(self.page2.metadatas.count(), 2)
        metadata = self.page1.metadatas.get(value='Some entity 1')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertIsNone(metadata.entity)
        metadata = self.page2.metadatas.get(value='Some entity 2')
        self.assertEqual(metadata.type, MetaType.Text)
        self.assertIsNotNone(metadata.entity)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.folder1.id),
            'corpus': str(self.corpus.id),
            'transcriptions': True,
            'elements': True,
            'entities': True,
            'drop': False,
        })
