from unittest.mock import patch
from django.db.models import Q
from arkindex_common.enums import TranscriptionType, MetaType, EntityType
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.consumers import ReindexConsumer
from arkindex.documents.models import Corpus, Element, Transcription, Entity, DataSource


@patch('arkindex.documents.consumers.Indexer')
class TestReindexConsumer(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        source = DataSource.objects.get(slug='test')

        cls.folder = cls.corpus.elements.get(name='Volume 1')
        cls.folder.metadatas.create(
            type=MetaType.Text,
            name='something',
            value='Some entity',
            entity=cls.corpus.entities.create(
                type=EntityType.Person,
                name='Some entity',
                source=source,
            )
        )

        corpus2 = Corpus.objects.create(name='Another corpus')
        element2 = corpus2.elements.create(
            type=corpus2.types.create(display_name='Element'),
            name='An element',
        )
        ts = element2.transcriptions.create(
            score=0.8,
            text='something',
            type=TranscriptionType.Word,
            source=source,
        )
        ts.transcription_entities.create(
            entity=corpus2.entities.create(
                type=EntityType.Misc,
                name='Some other entity',
                source=source,
            ),
            offset=0,
            length=1,
        )

    def assertQuerysetEqual(self, queryset1, queryset2, **options):
        """
        Make Django's assertQuerysetEqual slightly nicer to use
        """
        options.setdefault('ordered', False)
        return super().assertQuerysetEqual(queryset1, map(repr, queryset2), **options)

    def _assert_all_elements(self, call_args):
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(queryset, Element.objects.all())
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_all_entities(self, call_args):
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(queryset, Entity.objects.all())
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_all_transcriptions(self, call_args):
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(queryset, Transcription.objects.all())
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_all(self, mock):
        self.assertEqual(mock().run_index.call_count, 3)
        elements_call, entities_call, ts_call = sorted(mock().run_index.call_args_list, key=repr)
        self._assert_all_elements(elements_call)
        self._assert_all_entities(entities_call)
        self._assert_all_transcriptions(ts_call)

    def test_reindex_all(self, mock):
        ReindexConsumer({}).reindex_start({})
        self.assertEqual(mock().drop_index.call_count, 0)
        self._assert_all(mock)

    def test_reindex_drop(self, mock):
        ReindexConsumer({}).reindex_start({
            'drop': True,
        })
        self.assertEqual(mock().drop_index.call_count, 3)
        mock().drop_index.assert_any_call(Element.es_document)
        mock().drop_index.assert_any_call(Entity.es_document)
        mock().drop_index.assert_any_call(Transcription.es_document)
        self._assert_all(mock)

    def test_reindex_only_transcriptions(self, mock):
        ReindexConsumer({}).reindex_start({
            'transcriptions': True,
            'entities': False,
            'elements': False,
        })
        self.assertEqual(mock().drop_index.call_count, 0)
        self.assertEqual(mock().run_index.call_count, 1)
        self._assert_all_transcriptions(mock().run_index.call_args)

    def test_reindex_only_elements(self, mock):
        ReindexConsumer({}).reindex_start({
            'transcriptions': False,
            'entities': False,
            'elements': True,
        })
        self.assertEqual(mock().drop_index.call_count, 0)
        self.assertEqual(mock().run_index.call_count, 1)
        self._assert_all_elements(mock().run_index.call_args)

    def test_reindex_only_entities(self, mock):
        ReindexConsumer({}).reindex_start({
            'transcriptions': False,
            'entities': True,
            'elements': False,
        })
        self.assertEqual(mock().drop_index.call_count, 0)
        self.assertEqual(mock().run_index.call_count, 1)
        self._assert_all_entities(mock().run_index.call_args)

    def test_reindex_corpus(self, mock):
        ReindexConsumer({}).reindex_start({
            'corpus': str(self.corpus.id),
        })
        self.assertEqual(mock().drop_index.call_count, 0)
        self.assertEqual(mock().run_index.call_count, 3)
        elements_call, entities_call, ts_call = sorted(mock().run_index.call_args_list, key=repr)

        (queryset, ), kwargs = elements_call
        self.assertQuerysetEqual(queryset, self.corpus.elements.all())
        self.assertDictEqual(kwargs, {'bulk_size': 100})

        (queryset, ), kwargs = entities_call
        self.assertQuerysetEqual(queryset, self.corpus.entities.all())
        self.assertDictEqual(kwargs, {'bulk_size': 400})

        (queryset, ), kwargs = ts_call
        self.assertQuerysetEqual(queryset, Transcription.objects.filter(element__corpus_id=self.corpus.id))
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def test_reindex_element(self, mock):
        ReindexConsumer({}).reindex_start({
            'element': str(self.folder.id),
        })
        self.assertEqual(mock().drop_index.call_count, 0)
        self.assertEqual(mock().run_index.call_count, 3)
        entities_call, ts_call, elements_call = sorted(mock().run_index.call_args_list, key=repr)

        elements_list = list(Element.objects.get_descending(self.folder.id))
        elements_list.append(self.folder)

        (queryset, ), kwargs = elements_call
        self.assertQuerysetEqual(queryset, elements_list)
        self.assertDictEqual(kwargs, {'bulk_size': 100})

        (queryset, ), kwargs = entities_call
        self.assertQuerysetEqual(
            queryset,
            Entity.objects.filter(
                Q(metadatas__element__in=elements_list)
                | Q(transcriptions__element__in=elements_list)
            ),
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

        (queryset, ), kwargs = ts_call
        self.assertQuerysetEqual(queryset, Transcription.objects.filter(element__in=elements_list))
        self.assertDictEqual(kwargs, {'bulk_size': 400})
