from arkindex.documents.models import Corpus, AllowedMetaData
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import MetaType
from arkindex.users.models import User
from rest_framework import status
from django.urls import reverse
from unittest.mock import patch


class TestMetaData(FixtureAPITestCase):
    """
    Tests related to Metadata and creation limitations from API
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.admin = User.objects.get(is_admin=True)
        cls.internal = User.objects.create_internal_user('1337@te.st')
        cls.internal.verified_email = True
        cls.internal.save()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_vol = cls.private_corpus.elements.create(type=cls.vol.type, name='Vol')
        cls.allowed_meta = AllowedMetaData.objects.bulk_create(
            AllowedMetaData(corpus=cls.corpus, type=t, name=n)
            for t, n in (
                (MetaType.Date, 'edition'),
                (MetaType.Text, 'edition'),
            )
        )

    def setUp(self):
        super().setUp()
        self.metadata = self.vol.metadatas.create(type=MetaType.Text, name='folio', value='123')
        self.private_metadata = self.private_vol.metadatas.create(type=MetaType.HTML, name='leet', value='1337')

    def test_metadata_forbidden_methods(self):
        """
        Only POST method is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    @patch('arkindex.documents.api.elements.AllowedMetaDataPagination')
    def test_list_allowed_metadata(self, pagination_mock):
        pagination_mock.page_size = 3
        response = self.client.get(
            reverse('api:corpus-allowed-metadata', kwargs={'pk': str(self.corpus.id)}),
            data={'page_size': 3}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 5)
        self.assertListEqual(
            data['results'],
            [
                {'type': 'date', 'name': 'date'},
                {'type': 'date', 'name': 'edition'},
                {'type': 'location', 'name': 'location'}
            ]
        )
        response = self.client.get(data['next'])
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()['results'],
            [
                {'type': 'text', 'name': 'edition'},
                {'type': 'text', 'name': 'folio'}
            ]
        )

    def test_allowed_metadata_methods(self):
        """
        Allowed metadata is read-only endpoint
        """
        methods = (self.client.post, self.client.patch, self.client.put, self.client.delete)
        for method in methods:
            response = method(reverse('api:corpus-allowed-metadata', kwargs={'pk': str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_metadata_verified(self):
        response = self.client.post(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_metadata_methods(self):
        self.client.force_login(self.user)
        methods = (self.client.get, self.client.patch, self.client.put)
        for method in methods:
            response = method(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_metadata_writable_element(self):
        self.client.force_login(self.user)
        hidden_book = self.private_corpus.elements.create(
            type=self.corpus.types.get(slug='volume'),
            name='Nope'
        )
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(hidden_book.id)}),
            data={'type': 'text', 'name': 'language', 'value': 'Japanese'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_metadata(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Location, name='location')
        self.assertEqual(md.value, 'Texas')

    def test_create_metadata_date(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'date', 'value': '1885'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        metadata_id = self.vol.metadatas.get(type=MetaType.Date, name='date').id
        self.assertDictEqual(response.json(), {
            'id': str(metadata_id),
            'name': 'date',
            'revision': None,
            'type': 'date',
            'value': '1885',
            'dates': [{
                'day': None,
                'month': None,
                'type': 'exact',
                'year': 1885
            }]
        })

    def test_create_metadata_empty(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'name': ['This field is required.'],
            'type': ['This field is required.'],
            'value': ['This field is required.']
        })

    def test_create_metadata_index(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'text', 'name': 'edition', 'value': 'Hammersmith youth edition'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        metadata = self.vol.metadatas.filter(name='edition')
        self.assertEqual(metadata.count(), 2)
        self.assertEqual(metadata.get(type=MetaType.Date).index, 0)
        self.assertEqual(metadata.get(type=MetaType.Text).index, 1)

    def test_create_duplicated_metadata(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        metadata = self.vol.metadatas.filter(name='edition')
        self.assertEqual(metadata.count(), 1)

    def test_delete_metadata_verified(self):
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_metadata(self):
        self.client.force_login(self.user)
        self.assertEqual(self.vol.metadatas.count(), 1)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(self.vol.metadatas.count(), 0)

    def test_delete_non_allowed_metadata(self):
        """
        A user may not be able to delete a non allowed metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Edition of metadata with this (type, name) are disallowed for this corpus']}
        )

    def test_admin_delete_any(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.admin)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_metadata_private_corpus(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.private_metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_metadata_readable_element(self):
        """
        An explicit message should be raised when user can read but not delete metadata
        """
        self.client.force_login(self.user)
        self.private_corpus.corpus_right.create(user_id=self.user.id, can_write=False)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.private_metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have write access to this corpus.'})

    def test_admin_create_any(self):
        """
        Admin and internal users are allowed to create any metadata
        """
        for user in (self.admin, self.internal):
            self.client.force_login(user)
            response = self.client.post(
                reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                data={'type': 'text', 'name': 'certainly not allowed', 'value': 'bla bla bla'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
            md = self.vol.metadatas.get(type=MetaType.Text, name='certainly not allowed')
            self.assertEqual(md.value, 'bla bla bla')
            md.delete()

    def test_create_non_allowed_metadata(self):
        """
        Non internal user is limited in metadata creation
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'text', 'name': 'certainly not allowed', 'value': 'bla bla bla'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Metadata with this (type, name) are forbidden for this corpus']}
        )

    def test_patch_metadata(self):
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'date',
                'value': '2019-12-04'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'dates': [{'day': 4, 'month': 12, 'type': 'exact', 'year': 2019}],
            'id': str(self.metadata.id),
            'name': 'date',
            'revision': None,
            'type': 'date',
            'value': '2019-12-04'
        })
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.type, MetaType.Date)
        self.assertEqual(self.metadata.name, 'date')
        self.assertEqual(self.metadata.value, '2019-12-04')

    def test_patch_disallowed(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'death of William I, Count of Hainaut',
                'value': '1337'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Metadata with this (type, name) are forbidden for this corpus']}
        )

    def test_patch_from_any_internal(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.internal)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'I got you', 'value': 'patched !'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        metadata.refresh_from_db()
        self.assertEqual(metadata.value, 'patched !')

    def test_patch_to_any_internal(self):
        self.client.force_login(self.internal)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={'type': 'text', 'name': 'death of William I, Count of Hainaut', 'value': '1337'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.name, 'death of William I, Count of Hainaut')
        self.assertEqual(self.metadata.value, '1337')

    def test_patch_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'joke',
                'name': 'what is my name ?',
                'value': 'i have no name, im a machine'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'type': ['Value is not of type MetaType']})

    def test_patch_to_existing_metadata(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='111,222')
        self.assertEqual(metadata.index, 0)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'folio', 'value': '333,444'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        metadata.refresh_from_db()
        self.assertEqual(metadata.index, 1)
        self.assertEqual(metadata.value, '333,444')

    def test_patch_unallowed_metadata(self):
        """
        A user may not be able to patch a non allowed metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'folio', 'value': '333,444'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['Edition of metadata with this (type, name) are disallowed for this corpus']
        })
        metadata.refresh_from_db()
        self.assertEqual(metadata.name, 'not editable')
        self.assertEqual(metadata.value, '111,222')

    def test_get_metadata(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(self.metadata.id),
                'type': 'text',
                'name': 'folio',
                'value': '123',
                'revision': None,
                'dates': []
            }
        )
