from asyncmock import AsyncMock
from unittest.mock import patch
from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import MetaType, EntityType
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import DataSource, Transcription, Entity


class TestMLResults(FixtureTestCase):
    """
    Tests for ML results-related APIs
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.recognizer = DataSource.objects.get(name='Test Recognizer')
        cls.folder = cls.corpus.elements.get(name='Volume 1', type__folder=True)
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')

        ml_class = cls.corpus.ml_classes.create(name='Some class')
        region = cls.recognizer.regions.create(
            source=cls.recognizer,
            ml_class=ml_class,
            confidence=0.63,
            zone=cls.page.zone,
        )
        region.elements.add(cls.page)

        entity = cls.corpus.entities.create(
            name='An entity',
            type=EntityType.Misc,
            source=cls.recognizer,
        )
        entity2 = cls.corpus.entities.create(
            name='Another entity',
            type=EntityType.Misc,
            source=cls.recognizer,
        )

        cls.folder.classifications.create(ml_class=ml_class, source=cls.recognizer, confidence=0.42)
        cls.page.classifications.create(ml_class=ml_class, source=cls.recognizer, confidence=0.85)
        cls.page.metadatas.create(name='key', value='value', type=MetaType.Text, entity=entity)
        cls.page.transcriptions.first().transcription_entities.create(
            entity=entity2, offset=0, length=1,
        )

    def test_element_stats_requires_login(self):
        response = self.client.get(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_stats_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_stats(self):
        self.client.force_login(self.superuser)
        self.assertEqual(self.page.transcriptions.count(), 3)
        self.assertEqual(self.page.classifications.count(), 1)
        self.assertEqual(self.page.regions.count(), 1)
        self.assertEqual(Entity.objects.filter(transcriptions__element=self.page).count(), 1)
        self.assertEqual(Entity.objects.filter(metadatas__element=self.page).count(), 1)

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': str(self.recognizer.id),
                'type': 'recognizer',
                'slug': 'test',
                'name': 'Test Recognizer',
                'revision': '4.2',
                'internal': False,
                'classifications_count': 1,
                'transcriptions_count': 3,
                'regions_count': 1,
                'entities_count': 2,
            }
        ])

    def test_element_stats_folder(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:element-ml-stats', kwargs={'pk': str(self.folder.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': str(self.recognizer.id),
                'type': 'recognizer',
                'slug': 'test',
                'name': 'Test Recognizer',
                'revision': '4.2',
                'internal': False,
                'classifications_count': 2,
                'transcriptions_count': 9,
                'regions_count': 1,
                'entities_count': 2,
            }
        ])

    def test_corpus_stats_requires_login(self):
        response = self.client.get(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_stats_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_stats(self):
        self.client.force_login(self.superuser)
        self.assertEqual(self.folder.classifications.count(), 1)
        self.assertEqual(self.page.classifications.count(), 1)
        self.assertEqual(self.page.regions.count(), 1)
        self.assertEqual(Entity.objects.filter(transcriptions__element=self.page).count(), 1)
        self.assertEqual(Entity.objects.filter(metadatas__element=self.page).count(), 1)
        self.assertEqual(Transcription.objects.filter(element__corpus=self.corpus).count(), 9)

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': str(self.recognizer.id),
                'type': 'recognizer',
                'slug': 'test',
                'name': 'Test Recognizer',
                'revision': '4.2',
                'internal': False,
                'classifications_count': 2,
                'transcriptions_count': 9,
                'regions_count': 1,
                'entities_count': 2,
            }
        ])

    def test_corpus_destroy_results_requires_login(self):
        response = self.client.delete(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_destroy_results_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_corpus_destroy_results(self, get_layer_mock):
        get_layer_mock().send = AsyncMock()
        self.client.force_login(self.superuser)
        response = self.client.delete(reverse('api:corpus-ml-stats', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        get_layer_mock().send.assert_called_once_with('ml_results', {
            'type': 'ml_results.delete',
            'corpus_id': str(self.corpus.id),
            'element_id': None,
            'batch_size': 1000,
        })

    def test_element_destroy_results_requires_login(self):
        response = self.client.delete(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_destroy_results_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_element_destroy_results(self, get_layer_mock):
        get_layer_mock().send = AsyncMock()
        self.client.force_login(self.superuser)
        response = self.client.delete(reverse('api:element-ml-stats', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        get_layer_mock().send.assert_called_once_with('ml_results', {
            'type': 'ml_results.delete',
            'corpus_id': None,
            'element_id': str(self.page.id),
            'batch_size': 1000,
        })
