from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import ClassificationState, DataSource, MLClass, Element, Corpus, Classification, \
    MLToolType
from arkindex.project.tests import FixtureAPITestCase


class TestClasses(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name='text', corpus=cls.corpus)
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.element = Element.objects.get(name='Volume 1, page 1v')
        cls.folder = cls.corpus.elements.get(name='Volume 1')

    def _create_classification(self):
        return self.element.classifications.create(
            source=DataSource.objects.get(slug='test'),
            ml_class=self.text,
            confidence=.5,
        )

    def test_classification_creation(self):
        """
        Ensure classification creation works and set auto fields correctly
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:classification-create'), {
            'element': str(self.element.id),
            'ml_class': str(self.text.id)
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['source']['type'], 'classifier')
        self.assertEqual(response.data['source']['slug'], 'manual')
        self.assertDictEqual(response.data['ml_class'], {
            'id': str(self.text.id),
            'name': 'text'
        })
        self.assertEqual(response.data['state'], ClassificationState.Validated.value)
        self.assertEqual(response.data['confidence'], 1)
        self.assertEqual(response.data['high_confidence'], True)

    def test_classification_exists(self):
        """
        If a classification exists, creation must respond a 400_BAD_REQUEST
        with an explicit message
        """
        self.client.force_login(self.user)
        request = (
            reverse('api:classification-create'),
            {
                'element': str(self.element.id),
                'ml_class': str(self.text.id)
            }
        )
        response = self.client.post(*request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(*request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['A manual classification already exists for this `element` and this `class`']
        })

    def test_classification_ignored_params(self):
        """
        Ensure confidence and high_confidence attributes cannot be changed
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:classification-create'), {
            'element': str(self.element.id),
            'ml_class': str(self.text.id),
            'confidence': 0.5,
            'high_confidence': False
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertEqual(data['confidence'], 1)
        self.assertEqual(data['high_confidence'], True)

    def test_classification_creation_without_permission(self):
        """
        Without permission, should not raise a 403 Forbidden: indicates that the resource exists
        Indicates that page PK does not exist instead
        """
        response = self.client.post(reverse('api:classification-create'), {
            'element': str(self.element.id),
            'ml_class': str(self.text.id)
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.data, {
            'element': [
                'Invalid pk "{!s}" - object does not exist.'.format(self.element.id),
            ]
        })

    def test_classification_validate(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'name': classification.source.name,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_classification_validate_without_permissions(self):
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_classification_reject(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'name': classification.source.name,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_classification_reject_without_permissions(self):
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_classification_can_still_be_moderated(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        classification.moderator = self.user
        classification.state = ClassificationState.Validated.value
        classification.save()

        # First try to reject
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'name': classification.source.name,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

        # Then try to validate
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'name': classification.source.name,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence,
            'high_confidence': False
        })

    def test_classification_selection_requires_login(self):
        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_classification_private_corpus(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.private_corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_classification_wrong_corpus(self):
        line = MLClass.objects.create(name='line', corpus=self.private_corpus)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'ml_class': line.id, 'mode': 'create'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': ['Ml class {} does not belong to the corpus {}'.format(
            line,
            self.corpus
        )]})

    def test_classification_wrong_ml_class(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'ml_class': 'ffffffff-ffff-ffff-ffff-ffffffffffff', 'mode': 'create'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'ml_class': ['Ml class ffffffff-ffff-ffff-ffff-ffffffffffff not found']
        })

    def test_classifications_wrong_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'test'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_classification_selection_create(self):
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act xxx',
        )
        act_y = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act yyy',
        )

        self.assertFalse(self.folder.classifications.exists())
        self.assertFalse(self.element.classifications.exists())

        self.client.force_login(self.superuser)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.folder.id, self.element.id, act_x.id, act_y.id]},
            format='json'
        )
        response = self.client.get(reverse('api:elements-selection') + '?best_classes=true')
        results = response.json()['results']
        self.assertEqual(len(results), 4)

        self.assertCountEqual(
            [element['id'] for element in results],
            [str(self.element.id), str(self.folder.id), str(act_x.id), str(act_y.id)],
        )

        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'ml_class': self.text.id, 'mode': 'create'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(self.folder.classifications.exists())
        classification = self.folder.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertEqual(classification.source.slug, 'manual')
        self.assertEqual(classification.confidence, 1)

        self.assertTrue(self.element.classifications.exists())
        classification = self.element.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertEqual(classification.source.slug, 'manual')
        self.assertEqual(classification.confidence, 1)

    def test_classifications_selection_validate(self):
        source_1 = DataSource.objects.get(slug='test')
        source_2 = DataSource.objects.create(
            type=MLToolType.NER,
            slug='test',
            name='Test NER',
            revision='4.2',
            internal=False,
        )

        line = MLClass.objects.create(name='line', corpus=self.private_corpus)
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name='Act xxx',
        )
        Classification.objects.create(
            element=act_x,
            source=source_1,
            state=ClassificationState.Pending,
            high_confidence=True,
            ml_class=line
        )

        for e in [self.folder, self.element]:
            Classification.objects.create(
                element=e,
                source=source_1,
                state=ClassificationState.Pending,
                high_confidence=True,
                ml_class=self.text
            )
            Classification.objects.create(
                element=e,
                source=source_2,
                state=ClassificationState.Pending,
                high_confidence=False,
                ml_class=self.text
            )

        self.client.force_login(self.superuser)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.folder.id, self.element.id, act_x.id]},
            format='json'
        )

        response = self.client.post(
            reverse('api:classification-selection'),
            data={'corpus_id': self.corpus.id, 'mode': 'validate'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        for e in [self.folder, self.element]:
            classification = e.classifications.get(state=ClassificationState.Validated)
            self.assertTrue(classification.high_confidence)
            self.assertEqual(classification.source, source_1)

            classification = e.classifications.get(state=ClassificationState.Pending)
            self.assertFalse(classification.high_confidence)
            self.assertEqual(classification.source, source_2)

        classification = act_x.classifications.get()
        self.assertEqual(classification.state, ClassificationState.Pending)
