from datetime import timedelta
from math import ceil
from timeit import default_timer
from django.core.management.base import BaseCommand
from arkindex.project.polygon import Polygon
from arkindex.images.models import Zone
import logging

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
)
logger = logging.getLogger()


class Command(BaseCommand):
    help = 'Sort points inside zone polygons in bulk'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--offset',
            type=int,
            default=0,
        )
        parser.add_argument(
            '--limit',
            type=int,
            default=0,
        )
        parser.add_argument(
            '--batch-size',
            help='Batch size used to retrieve zones.',
            type=int,
            default=1000,
        )
        parser.add_argument(
            '--update-batch-size',
            help='Batch size used to update zones. Updates are run with each retrieval batch.',
            type=int,
            default=1000,
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            default=False,
        )
        parser.add_argument(
            '--no-remove',
            action='store_false',
            dest='remove',
            default=True,
        )

    def handle(self,
               *args,
               offset=0,
               limit=0,
               batch_size=1000,
               update_batch_size=1000,
               dry_run=False,
               remove=True,
               **kwargs):
        assert batch_size > 0, 'Batch size must be positive.'
        assert update_batch_size > 0, 'Update batch size must be positive.'

        assert offset >= 0, 'Offset cannot be negative.'
        queryset = Zone.objects.order_by('id')
        if limit:
            assert limit > 0, 'Limit must be positive.'
            queryset = queryset[offset:offset + limit]
        else:
            queryset = queryset[offset:]

        completed, reordered, failed, removed, total = 0, 0, 0, 0, queryset.count()
        logger.info('{} zones to reorder ({} batches)'.format(total, ceil(total / batch_size)))
        start_time = default_timer()
        for i in range(0, total, batch_size):
            batch = queryset[i:min(i + batch_size, total)]
            reordered_zones = []

            for zone in batch:
                old_polygon = Polygon(zone.polygon, reorder=False)

                try:
                    zone.polygon.reorder()
                except Exception as e:
                    if old_polygon.width < 1 or old_polygon.height < 1 and remove:
                        if dry_run:
                            logger.info('Would remove incoherent zone {}'.format(zone.id))
                        else:
                            logger.info('Removing incoherent zone {}…'.format(zone.id))
                            zone.transcriptions.all().delete()
                            zone.elements.all().delete()
                            zone.regions.all().delete()
                            zone.delete()
                        removed += 1
                        total -= 1
                    else:
                        logger.warning('Polygon reordering failed on zone {!s} with polygon {!s}: {!s}'.format(
                            zone.id, old_polygon, e))
                        failed += 1
                    continue

                if zone.polygon != old_polygon:
                    # Check if a zone already exists and either reorder or deduplicate
                    existing_zone = Zone.objects \
                        .filter(image_id=zone.image_id, polygon=zone.polygon) \
                        .exclude(id=zone.id) \
                        .first()
                    if not existing_zone:
                        reordered_zones.append(zone)
                        continue

                    if dry_run:
                        logger.info('Would remove zone {} and reassign to zone {}'.format(zone.id, existing_zone.id))
                        continue

                    logger.info('Removing zone {} and reassigning to zone {}…'.format(zone.id, existing_zone.id))
                    zone.transcriptions.all().update(zone_id=existing_zone.id)
                    zone.elements.all().update(zone_id=existing_zone.id)
                    zone.regions.all().update(zone_id=existing_zone.id)
                    zone.delete()
                    removed += 1
                    total -= 1

            if not dry_run:
                Zone.objects.bulk_update(reordered_zones, ['polygon'], batch_size=update_batch_size)
            reordered += len(reordered_zones)
            completed += batch.count()

            percent = (completed / total) * 100
            elapsed = timedelta(seconds=default_timer() - start_time)
            logger.info('{:6.2f}% completed ({:7d} / {}) in {}, {} reordered, {} failed, {} removed'.format(
                percent, completed, total, elapsed, reordered, failed, removed
            ))
