from django.conf import settings
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.functions import Concat, Substr
from django.utils.functional import cached_property
from django.utils.text import slugify
from arkindex.dataimport.models import DataFile
from arkindex.images.managers import ImageServerManager
from arkindex.project.models import IndexableModel
from arkindex.project.fields import StripSlashURLField, LStripTextField, MD5HashField
from arkindex.project.polygon import PolygonField
from arkindex.project.aws import S3FileMixin, S3FileModelMixin, S3FileStatus
from io import BytesIO
from PIL import Image as PillowImage
import logging
import os
import requests
import urllib.parse

logger = logging.getLogger(__name__)
profile_uri_validator = URLValidator(schemes=['http', 'https'], message='Invalid IIIF profile URI')


class ImageServer(models.Model):
    """
    A IIIF image server
    """
    display_name = models.CharField(max_length=250)
    url = StripSlashURLField(unique=True)
    s3_bucket = models.SlugField(max_length=63, blank=True, null=True, db_index=False)
    s3_region = models.SlugField(max_length=63, blank=True, null=True, db_index=False)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    validated = models.BooleanField(default=False)
    read_only = models.BooleanField(default=False)

    objects = ImageServerManager()

    def __str__(self):
        return self.display_name

    @property
    def is_local(self):
        """
        Indicates whether the server is the Arkindex IIIF server.
        """
        return self.id == settings.LOCAL_IMAGESERVER_ID

    def find_image(self, path, offline=False, width=None, height=None, save=True):
        """
        Lookup an image on a server
        This is the preferred way to construct an image
        """
        img = None
        # Try direct access to path
        img = self.images.filter(path=path).first()

        if img is None:
            # Try the url encoded path
            img = self.images.filter(
                path=urllib.parse.quote_plus(path),
            ).first()

        if img is None:
            img = Image(server=self, path=path, width=width, height=height)

        # Support offline queries
        if not offline:
            img.perform_check(save=save)

        if save:
            img.save()

        return img

    def build_url(self, path):
        """
        Build a full url on this server
        """
        base_url = self.url
        path = path.lstrip('/')
        if not base_url.endswith('/'):
            base_url += '/'
        abs_url = urllib.parse.urljoin(base_url, path)

        # In some cases, the path will contain something that looks like a scheme,
        # e.g. ark:/something with Gallica URLs. In those cases, urljoin will not add
        # the server's root URL ; we fall back to a dirty way of joining the URLs.
        if abs_url.startswith(base_url):
            return abs_url

        return base_url + path  # Don't try this at home

    def merge_into(self, other):
        """
        Merge this ImageServer into another. This does not remove the ImageServer,
        it only moves around the images.
        This allows changing the root folder of a server and updating all of its images.
        Both servers must have the same URL schemes and domains.
        If both servers have the same root path, the images are simply moved.
        If the destination server is a parent folder of this server, the subfolder is prepended to all image paths.
        If the destination server is a child folder of this server, the subfolder is removed from all paths.
        In the latter case, merging is only allowed when all images of the server are in the same subfolder.
        Raises ValueError whenever merging is impossible.
        """
        assert isinstance(other, ImageServer)
        if not self.images.exists():
            return

        # Parse both URLs to first check for scheme/netloc compatibility
        src_url, dest_url = map(urllib.parse.urlsplit, (self.url, other.url))
        if src_url.scheme != dest_url.scheme:
            raise ValueError('Cannot merge into an image server of a different protocol')
        elif src_url.netloc != dest_url.netloc:
            raise ValueError('Cannot merge into an image server on a different domain')

        # Check paths
        src_path, dest_path = src_url.path.strip('/'), dest_url.path.strip('/')
        common = os.path.commonpath([src_path, dest_path])

        if common == dest_path:
            # Destination is a parent folder of this server
            prefix = src_path[len(common):].lstrip('/')
            if not prefix.endswith('/'):
                prefix += '/'
            # Prepend all image paths with this new prefix and change their server
            self.images.update(
                server_id=other.id,
                path=Concat(models.Value(prefix), 'path'),
            )
            return

        if common == src_path:
            # Destination is a child of this server
            prefix = dest_path[len(common):].lstrip('/')
            if not prefix.endswith('/'):
                prefix += '/'
            # Check that all images on this server start with the prefix
            if self.images.exclude(path__startswith=prefix).exists():
                raise ValueError('Cannot merge into an image server on a child folder: '
                                 'Some images are not in the child folder')
            # Remove the prefix from all image paths and change their server
            # Note that positions in Substr() are 1-indexed
            self.images.update(
                server_id=other.id,
                path=Substr('path', len(prefix) + 1),
            )
            return

        raise ValueError('Cannot merge into an image server on a different subfolder')

    def split(self, folder):
        """
        Create a new ImageServer for a subfolder on this server and move the images there.
        Raises ValueError if there are no images to move.
        Returns the newly created ImageServer.
        """
        # Ensure we have a path ending with a single slash
        folder_slash = folder.lstrip('/') + '/'
        images = self.images.filter(path__startswith=folder_slash)
        if not images.exists():
            raise ValueError('No images are in the "{}" folder'.format(folder))

        # Create the new server on the subfolder
        new_server = ImageServer.objects.create(
            display_name='{}_{}'.format(self.display_name, slugify(folder)),
            url=urllib.parse.urljoin(self.url.lstrip('/') + '/', folder),
        )

        # Move the images there, removing the subfolder from their path
        images.update(
            server_id=new_server.id,
            # Substr positions are 1-indexed
            path=Substr('path', len(folder_slash) + 1)
        )
        return new_server

    @property
    def grafana_tag(self):
        return '{}_{}'.format(self.display_name, self.id)


class Image(S3FileModelMixin, IndexableModel):
    """
    A document image
    """
    server = models.ForeignKey(ImageServer, on_delete=models.CASCADE, related_name='images')
    path = LStripTextField(chars='/')
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)
    datafile = models.ForeignKey(DataFile, related_name='images', null=True, on_delete=models.SET_NULL)
    # Overwrite s3 hash to allow null value for external servers
    hash = MD5HashField(blank=True, null=True)

    @cached_property
    def s3_bucket(self):
        return self.server.s3_bucket

    @cached_property
    def s3_region(self):
        return self.server.s3_region

    @cached_property
    def s3_key(self):
        return urllib.parse.unquote(self.path)

    class Meta:
        unique_together = (
            ('server', 'path'),
            ('server', 'hash'),
        )

    @property
    def url(self):
        return self.server.build_url(self.path)

    def get_thumbnail_url(self, max_width=200, max_height=None):
        if max_width is None and max_height is None:
            resize = "full"
        else:
            resize = "{0},{1}".format(max_width or '', max_height or '')
        return "{0}/full/{1}/0/default.jpg".format(self.url.rstrip('/'), resize)

    @property
    def s3_put_url(self):
        assert self.server.is_local, 'Cannot PUT images on remote servers'
        return super(S3FileModelMixin, self).s3_put_url

    def perform_check(self, save=True, raise_exc=False):
        """
        Check the image's existence and update width, height and status properties
        """
        url = self.url
        if not url.endswith('/'):
            url += '/'
        info_url = urllib.parse.urljoin(url, 'info.json')
        try:
            # Load info
            resp = requests.get(info_url, timeout=15, allow_redirects=True)
            resp.raise_for_status()

            # Load data
            data = resp.json()
            assert all(item in data for item in ('@id', 'width', 'height', 'profile')), 'Missing required properties'
            assert isinstance(data['profile'], (str, list)), 'Profile is not a list or URI'

            # Check id
            image_id = data['@id']
            assert image_id.startswith(self.server.url), \
                'Image id does not start with server url ({} vs. {})'.format(image_id, self.server.url)
        except Exception as e:
            logger.warn('Image check failed: {}'.format(str(e)))
            self.status = S3FileStatus.Error
            if save:
                self.save()
            if raise_exc:
                raise
            return

        # Ensure that the required profile URI is a URI
        if isinstance(data['profile'], list):
            profile_uri = data['profile'][0]
        else:
            profile_uri = data['profile']
        try:
            profile_uri_validator(profile_uri)
        except ValidationError as e:
            logger.warning(', '.join(list(e)))

        # Use image resource ID from IIIF server to update the image path if needed
        self.path = image_id[len(self.server.url) + 1:]
        self.width, self.height = int(data['width']), int(data['height'])

        # Handle optional maxWidth and maxHeight properties in the profile description
        # In the IIIF Image API 2.x, `profile` is a list with one or more Image API compliance level URIs,
        # and optionally an object, the profile description.
        # In 1.x, `profile` is a single URI, not a list, and 1.x does not have maximum size properties.
        max_width, max_height = None, None
        if isinstance(data['profile'], list) and len(data['profile']) >= 2:
            profile_desc = next((item for item in data['profile'] if isinstance(item, dict)), {})
            max_width = int(profile_desc.get('maxWidth', 0))
            max_height = int(profile_desc.get('maxHeight', 0))

        # Special case for Harvard IDS: those properties are wrongly defined in the main object
        if not max_width and 'maxWidth' in data:
            max_width = int(data.get('maxWidth', 0))
        if not max_height and 'maxHeight' in data:
            max_height = int(data.get('maxHeight', 0))

        # From the Image API specification: If maxWidth is specified and maxHeight is not,
        # then clients should infer that maxHeight = maxWidth.
        if max_width and not max_height:
            max_height = max_width

        # Update the image size to reflect the maximum allowed size
        if (max_width or max_height) and (self.width > max_width or self.height > max_height):
            ratio = min(max_width / self.width, max_height / self.height)
            self.width = round(ratio * self.width)
            self.height = round(ratio * self.height)

        self.status = S3FileStatus.Checked
        if save:
            self.save()

    def pillow_open(self, max_width=500):
        if self.server.is_local:
            b = BytesIO()
            self.s3_object.download_fileobj(b)
            return PillowImage.open(b)

        # PIL.Image.open explicitly requires seek(int) method, that the urllib responses do not provide.
        # We therefore have to get the whole content and put it back in a file-like object
        resp = requests.get(
            self.get_thumbnail_url(max_width=max_width),
            timeout=settings.IIIF_DOWNLOAD_TIMEOUT,
        )
        resp.raise_for_status()
        return PillowImage.open(BytesIO(resp.content))

    def pillow_save(self, pillow_img, format=None):
        """
        Save a PIL.Image instance to S3
        """
        assert self.server.is_local, 'Only images on the local server can be saved'
        logger.debug('Saving {} to S3'.format(self.id))
        b = BytesIO()
        pillow_img.save(b, format=format)
        b.seek(0)
        self.s3_object.upload_fileobj(b)

    def __str__(self):
        return '{} - {}'.format(self.id, self.url)


class Thumbnail(S3FileMixin):
    """
    Describes an element thumbnail
    """
    s3_bucket = settings.AWS_THUMBNAIL_BUCKET

    def __init__(self, element):
        self.element = element

    @property
    def name(self):
        return '{}.jpg'.format(str(self.element.id.hex))

    @cached_property
    def s3_key(self):
        return self.name


class Zone(IndexableModel):
    """
    A zone on an image
    """
    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='zones')
    polygon = PolygonField()

    class Meta:
        constraints = [
            # Handled using polygon_hash, a PostgreSQL generated column that performs
            # a MD5 hash of the polygon.
            models.UniqueConstraint(
                fields=['image', 'polygon'],
                name='zone_unique_image_polygon',
            ),
        ]

    def __str__(self):
        return 'Zone {}'.format(self.polygon)

    @property
    def url(self):
        '''
        Build the IIIF crop url
        '''
        return urllib.parse.urljoin(
            self.image.url + '/',
            '{},{},{},{}/full/0/default.jpg'.format(
                self.polygon.x, self.polygon.y, self.polygon.width, self.polygon.height),
        )
