"""
Django settings for arkindex project.

Generated by 'django-admin startproject' using Django 1.11.6.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import logging
import os
import sys
import warnings
from datetime import timedelta
from corsheaders.defaults import default_headers


def env2list(env_name, separator=',', default=[]):
    '''
    Load env variable as a list
    '''
    value = os.environ.get(env_name)
    return value and value.split(separator) or default


# Database
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.environ.get('DB_NAME', 'arkindex_dev'),
        'USER': os.environ.get('DB_USER', 'devuser'),
        'PASSWORD': os.environ.get('DB_PASSWORD', 'devdata'),
        'HOST': os.environ.get('DB_HOST', 'localhost'),
        'PORT': os.environ.get('DB_PORT', 9100),
    }
}

# Admins in charge
ADMINS = [
    ('Bastien', 'abadie@teklia.com'),
    ('Erwan', 'rouchet@teklia.com'),
]

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
LOGS_DIR = os.environ.get('LOGS_DIR', os.path.realpath(os.path.join(BASE_DIR, '../logs')))
if not os.path.isdir(LOGS_DIR):
    os.makedirs(LOGS_DIR)
ML_CLASSIFIERS_DIR = os.environ.get('ML_CLASSIFIERS_DIR', os.path.join(BASE_DIR, '../../ml-classifiers'))

# Local IIIF server
LOCAL_IMAGESERVER_ID = int(os.environ.get('LOCAL_IMAGESERVER_ID', 1))

# Amazon S3
PONOS_AWS_ACCESS_KEY = AWS_ACCESS_KEY = os.environ.get('AWS_ACCESS_KEY_ID')
PONOS_AWS_SECRET_KEY = AWS_SECRET_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY')
PONOS_AWS_ENDPOINT = AWS_ENDPOINT = os.environ.get('AWS_ENDPOINT')
PONOS_AWS_REGION = AWS_REGION = os.environ.get('AWS_REGION')
PONOS_S3_LOGS_BUCKET = os.environ.get('PONOS_S3_LOGS_BUCKET', 'ponos-logs')
PONOS_S3_ARTIFACTS_BUCKET = os.environ.get('PONOS_S3_ARTIFACTS_BUCKET', 'ponos-artifacts')
AWS_THUMBNAIL_BUCKET = os.environ.get('AWS_THUMBNAIL_BUCKET', 'thumbnails')
AWS_STAGING_BUCKET = os.environ.get('AWS_STAGING_BUCKET', 'staging')

# Ponos integration
if os.environ.get('PONOS_TASK'):
    # In a ponos docker task
    PONOS_RECIPE = None
    ML_CLASSIFIERS_DIR = '/arkindex/classifiers'

else:
    # As scheduler or dev
    PONOS_RECIPE = {
        'env': {
            'ARKINDEX_API_URL': os.environ.get('ARKINDEX_API_URL', 'http://localhost:8000/api/v1/'),
            'ARKINDEX_API_TOKEN': os.environ.get('ARKINDEX_API_TOKEN', 'deadbeefTestToken'),
        },
    }
    PONOS_PRIVATE_KEY = os.environ.get('PONOS_PRIVATE_KEY_PATH', os.path.join(BASE_DIR, 'ponos.key'))

if 'test' in sys.argv:
    # Overrides for unit tests
    AWS_ACCESS_KEY = 'test'
    AWS_SECRET_KEY = 'test'
    AWS_ENDPOINT = 'http://s3'
    PONOS_PRIVATE_KEY = None

    # Turn Django's UnorderedObjectListWarning into exceptions
    warnings.filterwarnings('error', category=RuntimeWarning, module='django.core.paginator')
    warnings.filterwarnings('error', category=RuntimeWarning, module='rest_framework.pagination')

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('SECRET_KEY', 'jf0w^y&ml(caax8f&a1mub)(js9(l5mhbbhosz3gi+m01ex+lo')

# SECURITY WARNING: don't run with debug turned on in production!
ARKINDEX_ENV = os.environ.get('ARKINDEX_ENV', 'dev')
DEBUG = ARKINDEX_ENV == 'dev'

# Run all system checks when starting the server
ALL_CHECKS = os.environ.get('ALL_CHECKS') == 'true' or 'runserver' in sys.argv

ALLOWED_HOSTS = env2list('ALLOWED_HOSTS')

# Docker container name resolution
ALLOWED_HOSTS += ['127.0.0.1', 'localhost', 'backend', 'ark-backend']

# Required for django-debug-toolbar
INTERNAL_IPS = ['127.0.0.1', '127.0.1.1']

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.postgres',
    'django_admin_hstore_widget',

    # Tools
    'channels',
    'rest_framework',
    'rest_framework.authtoken',
    'corsheaders',
    'ponos',

    # Our apps
    'arkindex.images',
    'arkindex.documents',
    'arkindex.users',
    'arkindex.dataimport',
]

MIDDLEWARE = [
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'arkindex.project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'arkindex.project.wsgi.application'
ASGI_APPLICATION = 'arkindex.project.routing.application'


# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTH_USER_MODEL = 'users.User'

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/
STATIC_URL = '/static/'
STATIC_ROOT = os.environ.get('STATIC_ROOT')
STATICFILES_DIRS = []


# API
REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': (
        'arkindex.project.permissions.AllowAny',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
        'ponos.authentication.AgentAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'arkindex.project.pagination.PageNumberPagination',
    'DEFAULT_SCHEMA_CLASS': 'arkindex.project.openapi.AutoSchema',
    'PAGE_SIZE': 20,
}

SIMPLE_JWT = {
    'USER_ID_CLAIM': 'agent_id',
    'ROTATE_REFRESH_TOKENS': True,
    'ACCESS_TOKEN_LIFETIME': timedelta(hours=6),
    'SIGNING_KEY': os.environ.get('SIGNING_KEY', SECRET_KEY),
}

# Elastic search config
ELASTIC_SEARCH_HOSTS = [
    os.environ.get('ES_HOST', 'localhost'),
]
# The Scroll API is required to go over 10K results
ES_RESULTS_LIMIT = 10000
# ES defaults to three items returned in a nested query if the inner_hits size is not defined
ES_INNER_RESULTS_LIMIT = 6

# Silent logger for elasticsearch
logging.getLogger('elasticsearch').setLevel(logging.WARNING)

# InfluxDB API root
INFLUXDB_API_URL = os.environ.get('INFLUXDB_API_URL', 'http://localhost:8086/')

# SSLify proxy
SSLIFY_HOST = os.environ.get('SSLIFY_HOST')

# Use SSL proxy
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

# IIIF manifests
IIIF_PRESENTATION_CONTEXT = "http://iiif.io/api/presentation/2/context.json"
IIIF_IMAGE_CONTEXT = "http://iiif.io/api/image/2/context.json"
IIIF_SEARCH_CONTEXT = "http://iiif.io/api/search/0/context.json"
IIIF_IMAGE_SERVICE_PROFILE = "http://iiif.io/api/image/2/level2.json"
IIIF_SEARCH_SERVICE_PROFILE = "http://iiif.io/api/search/0/search"

# IIIF manifest download timeout
# See http://docs.python-requests.org/en/master/user/advanced/#timeouts
IIIF_DOWNLOAD_TIMEOUT = (30, 60)

# check_images sample size when checking all servers
CHECK_IMAGES_SAMPLE_SIZE = 20

TRANSCRIPTIONS_IMPORT_QUEUE_SIZE = 25000
TRANSCRIPTIONS_IMPORT_CHUNK_SIZE = 10000

# GitLab OAuth
GITLAB_APP_ID = os.environ.get('GITLAB_APP_ID')
GITLAB_APP_SECRET = os.environ.get('GITLAB_APP_SECRET')

# Cache into memcached
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': '{}:11211'.format(os.environ.get('CACHE_HOST', 'localhost')),
    }
}

# Django Channels layer using Redis
REDIS_HOST = os.environ.get('REDIS_HOST', 'localhost')
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [
                (REDIS_HOST, 6379)
            ],
        },
    },
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'console_debug': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'filters': ['require_debug_true'],
        },
        'importers': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'formatter': 'verbose',
            'filename': os.path.join(LOGS_DIR, 'importers.log'),
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
            'include_html': True,
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console', 'mail_admins', ],
            'level': 'INFO',
            'propagate': True,
        },
        'daphne.server': {
            'handlers': ['console_debug'],
            'level': 'INFO',
        },
        'arkindex.documents.consumers': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'elasticsearch.trace': {
            'handlers': ['console_debug'],
            'level': 'DEBUG',
            'propagate': False,
        },
    },
    'formatters': {
        'verbose': {
            'format': '[%(levelname)s] %(asctime)s %(module)s: %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
}

# Email
EMAIL_SUBJECT_PREFIX = '[Arkindex {}] '.format(ARKINDEX_ENV)
if os.environ.get('EMAIL_HOST'):
    EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
    EMAIL_HOST = os.environ.get('EMAIL_HOST')
    EMAIL_PORT = os.environ.get('EMAIL_PORT')
    EMAIL_HOST_USER = os.environ.get('EMAIL_HOST_USER')
    DEFAULT_FROM_EMAIL = SERVER_EMAIL = EMAIL_HOST_USER
    EMAIL_HOST_PASSWORD = os.environ.get('EMAIL_HOST_PASSWORD')
    EMAIL_USE_TLS = True
else:
    EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'


# Cookies
def samesite(name, default):
    env = '{}_COOKIE_SAMESITE'.format(name)
    value = os.environ.get(env, default).lower()
    if value == 'none':
        value = None
    assert value in ('lax', 'strict', None), 'Invalid {} value {}'.format(env, value)
    return value


CSRF_COOKIE_NAME = os.environ.get('CSRF_COOKIE_NAME', 'arkindex.csrf')
CSRF_TRUSTED_ORIGINS = env2list('CSRF_TRUSTED_ORIGINS')
CSRF_COOKIE_DOMAIN = os.environ.get('COOKIE_DOMAIN')
CSRF_COOKIE_SAMESITE = samesite('CSRF', 'lax')

SESSION_COOKIE_NAME = os.environ.get('SESSION_COOKIE_NAME', 'arkindex.auth')
SESSION_COOKIE_DOMAIN = os.environ.get('COOKIE_DOMAIN')
SESSION_COOKIE_SAMESITE = samesite('SESSION', 'lax')
# Required for authentication over websockets
SESSION_COOKIE_HTTPONLY = False

CORS_ORIGIN_WHITELIST = env2list('CORS_ORIGIN_WHITELIST', default=[
    'universalviewer.io',
    'localhost:8080',
    '127.0.0.1:8080',
    'localhost:5000',
    '127.0.0.1:5000',
])
CORS_ALLOW_CREDENTIALS = True
CORS_ALLOW_HEADERS = default_headers + (
    'cache-control',  # Allow the frontend to prevent caching some API requests
)
CORS_URLS_REGEX = r'^/(api|ponos)/.*$'

# Support CORS suffixes
cors_suffixes = env2list('CORS_SUFFIXES')
if cors_suffixes:
    CORS_ORIGIN_REGEX_WHITELIST = [
        r"^https://.+{}".format(suffix)
        for suffix in cors_suffixes
    ]

# Docker images used by our ponos workflow
ARKINDEX_APP_IMAGE = os.environ.get('ARKINDEX_APP_IMAGE', 'arkindex-app')
ARKINDEX_TASKS_IMAGE = os.environ.get('ARKINDEX_TASKS_IMAGE', 'arkindex-tasks')

# User groups with special permissions
INTERNAL_GROUP_ID = 2

# Optional unit tests runner with code coverage
try:
    import django_nose # noqa
    INSTALLED_APPS.append('django_nose')
    TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'
    NOSE_ARGS = [
        '--with-coverage',
        '--cover-package=arkindex',
    ]
except ImportError:
    pass

# Local settings
if 'test' not in sys.argv:
    try:
        from .local_settings import * # noqa
    except ImportError:
        pass

try:
    import django_extensions  # noqa
    INSTALLED_APPS.append('django_extensions')
except ImportError:
    pass

try:
    import debug_toolbar # noqa
    MIDDLEWARE.insert(0, 'debug_toolbar.middleware.DebugToolbarMiddleware')
    INSTALLED_APPS.append('debug_toolbar')
except ImportError:
    pass
