"""
Helper methods to trigger tasks in asynchronous workers
"""
from typing import Union
from uuid import UUID
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from arkindex.documents.models import Element, Corpus

REINDEX_CHANNEL = 'reindex'
ML_RESULTS_CHANNEL = 'ml_results'
CORPUS_CHANNEL = 'corpus'

ACTION_REINDEX_START = 'reindex.start'
ACTION_ML_RESULTS_DELETE = 'ml_results.delete'
ACTION_CORPUS_DELETE = 'corpus.delete'


def reindex_start(*,
                  element: Union[Element, UUID, str] = None,
                  corpus: Union[Corpus, UUID, str] = None,
                  transcriptions: bool = False,
                  elements: bool = False,
                  entities: bool = False,
                  drop: bool = False) -> None:
    """
    Reindex elements into ElasticSearch.

    If `element` and `corpus` are left unspecified, all elements will be picked.
    If `element` is specified, its children will be included.

    `drop` will REMOVE all indexes entirely and recreate them before indexing.
    This is only allowed when reindexing all elements.
    """
    element_id = None
    corpus_id = None
    if isinstance(element, Element):
        element_id = str(element.id)
    elif element:
        element_id = str(element)

    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    elif corpus:
        corpus_id = str(corpus)

    async_to_sync(get_channel_layer().send)(REINDEX_CHANNEL, {
        'type': ACTION_REINDEX_START,
        'element': element_id,
        'corpus': corpus_id,
        'transcriptions': transcriptions,
        'elements': elements,
        'entities': entities,
        'drop': drop,
    })


def ml_results_delete(*,
                      element: Union[Element, UUID, str] = None,
                      corpus: Union[Corpus, UUID, str] = None,
                      batch_size: int = 1000) -> None:
    """
    Delete all ML results from all sources on a corpus
    or an element and its *direct* (non-recursive) children.
    """
    element_id = None
    corpus_id = None
    if isinstance(element, Element):
        element_id = str(element.id)
    elif element:
        element_id = str(element)

    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    elif corpus:
        corpus_id = str(corpus)

    assert element_id or corpus_id, 'Missing element or corpus ID'

    async_to_sync(get_channel_layer().send)(ML_RESULTS_CHANNEL, {
        'type': ACTION_ML_RESULTS_DELETE,
        'element_id': element_id,
        'corpus_id': corpus_id,
        'batch_size': batch_size,
    })


def corpus_delete(corpus: Union[Corpus, UUID, str],
                  *,
                  batch_size: int = 1000) -> None:
    """
    Delete a whole corpus without killing a server by removing all related
    models first and lowering the amount of foreign keys Django has to handle.
    """
    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    else:
        corpus_id = str(corpus)

    async_to_sync(get_channel_layer().send)(CORPUS_CHANNEL, {
        'type': ACTION_CORPUS_DELETE,
        'id': corpus_id,
        'batch_size': batch_size,
    })
